#!/usr/bin/env python
#-*- coding: utf-8 -*-
#
# Copyright 2013-2019 European Commission (JRC);
# Licensed under the EUPL (the 'Licence');
# You may not use this work except in compliance with the Licence.
# You may obtain a copy of the Licence at: http://ec.europa.eu/idabc/eupl
'''
wltp: generate WLTC gear-shifts based on vehicle characteristics
'''

from __future__ import division, unicode_literals

from ._version import (__version__, __updated__)


__copyright__     = "Copyright (C) 2013-2014 European Commission (JRC)"  # noqa
__license__       = "EUPL 1.1+"                     # noqa
__title__     = "wltp"                              # noqa
__summary__   = __doc__.splitlines()[0]             # noqa
__uri__       = "https://github.com/JRCSTU/wltp"  # noqa
